@@ -23,7 +23,7 @@
 //! - The encoding will be compatible with UTF-8. In particular, splitting an
 //!   encoded byte sequence by a UTF-8–encoded character always produces other
 //!   valid byte sequences. They can be re-encoded without error using
-//!   [`OsStrBytes::from_raw_bytes`] and similar methods.
+//!   [`RawOsString::into_os_string`] and similar methods.
 //!
 //! - All characters valid in platform strings are representable. [`OsStr`] and
 //!   [`OsString`] can always be losslessly reconstructed from extracted bytes.
@@ -73,14 +73,26 @@
 //!   For more information, see [`RawOsStr`][memchr complexity].
 //!
 //! - **raw\_os\_str** -
-//!   Enables use of [`RawOsStr`] and [`RawOsString`].
+//!   Provides:
+//!   - [`iter`]
+//!   - [`Pattern`]
+//!   - [`RawOsStr`]
+//!   - [`RawOsStrCow`]
+//!   - [`RawOsString`]
 //!
 //! ### Optional Features
 //!
-//! - **deprecated-byte-patterns** -
-//!   Provides an implementation of [`Pattern`] for [`u8`] that panics when it
-//!   is not an ASCII character. *This feature is deprecated and was only added
-//!   to ease removal of this pattern type.*
+//! - **checked\_conversions** -
+//!   Provides:
+//!   - [`EncodingError`]
+//!   - [`OsStrBytes::from_raw_bytes`]
+//!   - [`OsStringBytes::from_raw_vec`]
+//!   - [`RawOsStr::from_raw_bytes`]
+//!   - [`RawOsString::from_raw_vec`]
+//!
+//!   Because this feature should not be used in libraries, the
+//!   "OS_STR_BYTES_CHECKED_CONVERSIONS" environment variable must be defined
+//!   during compilation.
 //!
 //! - **print\_bytes** -
 //!   Provides implementations of [`print_bytes::ToBytes`] for [`RawOsStr`] and
@@ -103,21 +115,23 @@
 //!
 //! # Complexity
 //!
-//! The time complexities of trait methods will vary based on what
-//! functionality is available for the platform. At worst, they will all be
-//! linear, but some can take constant time. For example,
-//! [`OsStringBytes::from_raw_vec`] might be able to reuse the allocation for
-//! its argument.
+//! Conversion method complexities will vary based on what functionality is
+//! available for the platform. At worst, they will all be linear, but some can
+//! take constant time. For example, [`RawOsString::into_os_string`] might be
+//! able to reuse its allocation.
 //!
 //! # Examples
 //!
 //! ```
+//! # use std::io;
+//! #
+//! # #[cfg(feature = "raw_os_str")]
+//! # {
 //! # #[cfg(any())]
 //! use std::env;
 //! use std::fs;
-//! # use std::io;
 //!
-//! use os_str_bytes::OsStrBytes;
+//! use os_str_bytes::RawOsStr;
 //!
 //! # mod env {
 //! #   use std::env;
@@ -131,12 +145,13 @@
 //! # }
 //! #
 //! for file in env::args_os().skip(1) {
-//!     if file.to_raw_bytes().first() != Some(&b'-') {
+//!     if !RawOsStr::new(&file).starts_with('-') {
 //!         let string = "Hello, world!";
 //!         fs::write(&file, string)?;
 //!         assert_eq!(string, fs::read_to_string(file)?);
 //!     }
 //! }
+//! # }
 //! #
 //! # Ok::<_, io::Error>(())
 //! ```
@@ -151,6 +166,7 @@
 //! [sealed]: https://rust-lang.github.io/api-guidelines/future-proofing.html#c-sealed
 //! [print\_bytes]: https://crates.io/crates/print_bytes
 
+#![cfg_attr(not(feature = "checked_conversions"), allow(deprecated))]
 // Only require a nightly compiler when building documentation for docs.rs.
 // This is a private option that should not be used.
 // https://github.com/rust-lang/docs.rs/issues/147#issuecomment-389544407
@@ -161,6 +177,7 @@
     all(target_vendor = "fortanix", target_env = "sgx"),
     feature(sgx_platform)
 )]
+#![warn(unsafe_op_in_unsafe_fn)]
 #![warn(unused_results)]
 
 use std::borrow::Cow;
@@ -174,6 +191,41 @@ use std::path::Path;
 use std::path::PathBuf;
 use std::result;
 
+macro_rules! if_checked_conversions {
+    ( $($item:item)+ ) => {
+        $(
+            #[cfg(feature = "checked_conversions")]
+            $item
+        )+
+    };
+}
+
+#[cfg(not(os_str_bytes_docs_rs))]
+if_checked_conversions! {
+    const _: &str = env!(
+        "OS_STR_BYTES_CHECKED_CONVERSIONS",
+        "The 'OS_STR_BYTES_CHECKED_CONVERSIONS' environment variable must be \
+         defined to use the 'checked_conversions' feature.",
+    );
+}
+
+#[rustfmt::skip]
+macro_rules! deprecated_checked_conversion {
+    ( $message:expr , $item:item ) => {
+        #[cfg_attr(
+            not(feature = "checked_conversions"),
+            deprecated = $message
+        )]
+        $item
+    };
+}
+
+macro_rules! expect_encoded {
+    ( $result:expr ) => {
+        $result.expect("invalid raw bytes")
+    };
+}
+
 macro_rules! if_raw_str {
     ( $($item:item)+ ) => {
         $(
@@ -184,16 +236,24 @@ macro_rules! if_raw_str {
 }
 
 #[cfg_attr(
-    all(target_arch = "wasm32", target_os = "unknown"),
-    path = "wasm32/mod.rs"
+    all(target_family = "wasm", target_os = "unknown"),
+    path = "wasm/mod.rs"
 )]
 #[cfg_attr(windows, path = "windows/mod.rs")]
 #[cfg_attr(
-    not(any(all(target_arch = "wasm32", target_os = "unknown"), windows)),
+    not(any(all(target_family = "wasm", target_os = "unknown"), windows)),
     path = "common/mod.rs"
 )]
 mod imp;
 
+#[cfg(any(
+    all(
+        feature = "raw_os_str",
+        target_family = "wasm",
+        target_os = "unknown",
+    ),
+    windows,
+))]
 mod util;
 
 if_raw_str! {
@@ -204,32 +264,43 @@ if_raw_str! {
 
     mod raw_str;
     pub use raw_str::RawOsStr;
+    pub use raw_str::RawOsStrCow;
     pub use raw_str::RawOsString;
 }
 
-/// The error that occurs when a byte sequence is not representable in the
-/// platform encoding.
-///
-/// [`Result::unwrap`] should almost always be called on results containing
-/// this error. It should be known whether or not byte sequences are properly
-/// encoded for the platform, since [the module-level documentation][encoding]
-/// discourages using encoded bytes in interchange. Results are returned
-/// primarily to make panicking behavior explicit.
-///
-/// On Unix, this error is never returned, but [`OsStrExt`] or [`OsStringExt`]
-/// should be used instead if that needs to be guaranteed.
-///
-/// [encoding]: self#encoding
-/// [`OsStrExt`]: ::std::os::unix::ffi::OsStrExt
-/// [`OsStringExt`]: ::std::os::unix::ffi::OsStringExt
-/// [`Result::unwrap`]: ::std::result::Result::unwrap
-#[derive(Debug, Eq, PartialEq)]
-pub struct EncodingError(imp::EncodingError);
+deprecated_checked_conversion! {
+    "use `OsStrBytes::assert_from_raw_bytes` or \
+     `OsStringBytes::assert_from_raw_vec` instead, or enable the \
+     'checked_conversions' feature",
+    /// The error that occurs when a byte sequence is not representable in the
+    /// platform encoding.
+    ///
+    /// [`Result::unwrap`] should almost always be called on results containing
+    /// this error. It should be known whether or not byte sequences are
+    /// properly encoded for the platform, since [the module-level
+    /// documentation][encoding] discourages using encoded bytes in
+    /// interchange. Results are returned primarily to make panicking behavior
+    /// explicit.
+    ///
+    /// On Unix, this error is never returned, but [`OsStrExt`] or
+    /// [`OsStringExt`] should be used instead if that needs to be guaranteed.
+    ///
+    /// [encoding]: self#encoding
+    /// [`OsStrExt`]: ::std::os::unix::ffi::OsStrExt
+    /// [`OsStringExt`]: ::std::os::unix::ffi::OsStringExt
+    /// [`Result::unwrap`]: ::std::result::Result::unwrap
+    #[derive(Clone, Debug, Eq, PartialEq)]
+    #[cfg_attr(
+        os_str_bytes_docs_rs,
+        doc(cfg(feature = "checked_conversions"))
+    )]
+    pub struct EncodingError(imp::EncodingError);
+}
 
 impl Display for EncodingError {
     #[inline]
-    fn fmt(&self, formatter: &mut Formatter<'_>) -> fmt::Result {
-        self.0.fmt(formatter)
+    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
+        self.0.fmt(f)
     }
 }
 
@@ -237,6 +308,25 @@ impl Error for EncodingError {}
 
 type Result<T> = result::Result<T, EncodingError>;
 
+fn from_raw_bytes<'a, S>(
+    string: S,
+) -> result::Result<Cow<'a, OsStr>, imp::EncodingError>
+where
+    S: Into<Cow<'a, [u8]>>,
+{
+    match string.into() {
+        Cow::Borrowed(string) => imp::os_str_from_bytes(string),
+        Cow::Owned(string) => imp::os_string_from_vec(string).map(Cow::Owned),
+    }
+}
+
+fn cow_os_str_into_path(string: Cow<'_, OsStr>) -> Cow<'_, Path> {
+    match string {
+        Cow::Borrowed(string) => Cow::Borrowed(Path::new(string)),
+        Cow::Owned(string) => Cow::Owned(string.into()),
+    }
+}
+
 /// A platform agnostic variant of [`OsStrExt`].
 ///
 /// For more information, see [the module-level documentation][module].
@@ -244,14 +334,12 @@ type Result<T> = result::Result<T, EncodingError>;
 /// [module]: self
 /// [`OsStrExt`]: ::std::os::unix::ffi::OsStrExt
 pub trait OsStrBytes: private::Sealed + ToOwned {
-    /// Converts a byte slice into an equivalent platform-native string.
+    /// Converts a byte string into an equivalent platform-native string.
     ///
-    /// Provided byte strings should always be valid for the [unspecified
-    /// encoding] used by this crate.
+    /// # Panics
     ///
-    /// # Errors
-    ///
-    /// See documentation for [`EncodingError`].
+    /// Panics if the string is not valid for the [unspecified encoding] used
+    /// by this crate.
     ///
     /// # Examples
     ///
@@ -264,32 +352,70 @@ pub trait OsStrBytes: private::Sealed + ToOwned {
     ///
     /// let os_string = env::current_exe()?;
     /// let os_bytes = os_string.to_raw_bytes();
-    /// assert_eq!(os_string, OsStr::from_raw_bytes(os_bytes).unwrap());
+    /// assert_eq!(os_string, OsStr::assert_from_raw_bytes(os_bytes));
     /// #
     /// # Ok::<_, io::Error>(())
     /// ```
     ///
     /// [unspecified encoding]: self#encoding
-    fn from_raw_bytes<'a, S>(string: S) -> Result<Cow<'a, Self>>
+    #[must_use = "method should not be used for validation"]
+    #[track_caller]
+    fn assert_from_raw_bytes<'a, S>(string: S) -> Cow<'a, Self>
     where
         S: Into<Cow<'a, [u8]>>;
 
-    /// Converts a platform-native string into an equivalent byte slice.
+    deprecated_checked_conversion! {
+        "use `assert_from_raw_bytes` instead, or enable the \
+         'checked_conversions' feature",
+        /// Converts a byte string into an equivalent platform-native string.
+        ///
+        /// [`assert_from_raw_bytes`] should almost always be used instead. For
+        /// more information, see [`EncodingError`].
+        ///
+        /// # Errors
+        ///
+        /// See documentation for [`EncodingError`].
+        ///
+        /// # Examples
+        ///
+        /// ```
+        /// use std::env;
+        /// use std::ffi::OsStr;
+        /// # use std::io;
+        ///
+        /// use os_str_bytes::OsStrBytes;
+        ///
+        /// let os_string = env::current_exe()?;
+        /// let os_bytes = os_string.to_raw_bytes();
+        /// assert_eq!(os_string, OsStr::from_raw_bytes(os_bytes).unwrap());
+        /// #
+        /// # Ok::<_, io::Error>(())
+        /// ```
+        ///
+        /// [`assert_from_raw_bytes`]: Self::assert_from_raw_bytes
+        #[cfg_attr(
+            os_str_bytes_docs_rs,
+            doc(cfg(feature = "checked_conversions"))
+        )]
+        fn from_raw_bytes<'a, S>(string: S) -> Result<Cow<'a, Self>>
+        where
+            S: Into<Cow<'a, [u8]>>;
+    }
+
+    /// Converts a platform-native string into an equivalent byte string.
     ///
-    /// The returned bytes string will use an [unspecified encoding].
+    /// The returned string will use an [unspecified encoding].
     ///
     /// # Examples
     ///
     /// ```
-    /// use std::env;
-    /// # use std::io;
+    /// use std::ffi::OsStr;
     ///
     /// use os_str_bytes::OsStrBytes;
     ///
-    /// let os_string = env::current_exe()?;
-    /// println!("{:?}", os_string.to_raw_bytes());
-    /// #
-    /// # Ok::<_, io::Error>(())
+    /// let string = "foobar";
+    /// let os_string = OsStr::new(string);
+    /// assert_eq!(string.as_bytes(), &*os_string.to_raw_bytes());
     /// ```
     ///
     /// [unspecified encoding]: self#encoding
@@ -298,19 +424,20 @@ pub trait OsStrBytes: private::Sealed + ToOwned {
 }
 
 impl OsStrBytes for OsStr {
+    #[inline]
+    fn assert_from_raw_bytes<'a, S>(string: S) -> Cow<'a, Self>
+    where
+        S: Into<Cow<'a, [u8]>>,
+    {
+        expect_encoded!(from_raw_bytes(string))
+    }
+
     #[inline]
     fn from_raw_bytes<'a, S>(string: S) -> Result<Cow<'a, Self>>
     where
         S: Into<Cow<'a, [u8]>>,
     {
-        match string.into() {
-            Cow::Borrowed(string) => {
-                imp::os_str_from_bytes(string).map_err(EncodingError)
-            }
-            Cow::Owned(string) => {
-                OsStringBytes::from_raw_vec(string).map(Cow::Owned)
-            }
-        }
+        from_raw_bytes(string).map_err(EncodingError)
     }
 
     #[inline]
@@ -320,15 +447,20 @@ impl OsStrBytes for OsStr {
 }
 
 impl OsStrBytes for Path {
+    #[inline]
+    fn assert_from_raw_bytes<'a, S>(string: S) -> Cow<'a, Self>
+    where
+        S: Into<Cow<'a, [u8]>>,
+    {
+        cow_os_str_into_path(OsStr::assert_from_raw_bytes(string))
+    }
+
     #[inline]
     fn from_raw_bytes<'a, S>(string: S) -> Result<Cow<'a, Self>>
     where
         S: Into<Cow<'a, [u8]>>,
     {
-        OsStr::from_raw_bytes(string).map(|os_string| match os_string {
-            Cow::Borrowed(os_string) => Cow::Borrowed(Self::new(os_string)),
-            Cow::Owned(os_string) => Cow::Owned(os_string.into()),
-        })
+        OsStr::from_raw_bytes(string).map(cow_os_str_into_path)
     }
 
     #[inline]
@@ -344,14 +476,12 @@ impl OsStrBytes for Path {
 /// [module]: self
 /// [`OsStringExt`]: ::std::os::unix::ffi::OsStringExt
 pub trait OsStringBytes: private::Sealed + Sized {
-    /// Converts a byte vector into an equivalent platform-native string.
+    /// Converts a byte string into an equivalent platform-native string.
     ///
-    /// Provided byte strings should always be valid for the [unspecified
-    /// encoding] used by this crate.
+    /// # Panics
     ///
-    /// # Errors
-    ///
-    /// See documentation for [`EncodingError`].
+    /// Panics if the string is not valid for the [unspecified encoding] used
+    /// by this crate.
     ///
     /// # Examples
     ///
@@ -364,30 +494,66 @@ pub trait OsStringBytes: private::Sealed + Sized {
     ///
     /// let os_string = env::current_exe()?;
     /// let os_bytes = os_string.clone().into_raw_vec();
-    /// assert_eq!(os_string, OsString::from_raw_vec(os_bytes).unwrap());
+    /// assert_eq!(os_string, OsString::assert_from_raw_vec(os_bytes));
     /// #
     /// # Ok::<_, io::Error>(())
     /// ```
     ///
     /// [unspecified encoding]: self#encoding
-    fn from_raw_vec(string: Vec<u8>) -> Result<Self>;
+    #[must_use = "method should not be used for validation"]
+    #[track_caller]
+    fn assert_from_raw_vec(string: Vec<u8>) -> Self;
 
-    /// Converts a platform-native string into an equivalent byte vector.
+    deprecated_checked_conversion! {
+        "use `assert_from_raw_vec` instead, or enable the \
+         'checked_conversions' feature",
+        /// Converts a byte string into an equivalent platform-native string.
+        ///
+        /// [`assert_from_raw_vec`] should almost always be used instead. For
+        /// more information, see [`EncodingError`].
+        ///
+        /// # Errors
+        ///
+        /// See documentation for [`EncodingError`].
+        ///
+        /// # Examples
+        ///
+        /// ```
+        /// use std::env;
+        /// use std::ffi::OsString;
+        /// # use std::io;
+        ///
+        /// use os_str_bytes::OsStringBytes;
+        ///
+        /// let os_string = env::current_exe()?;
+        /// let os_bytes = os_string.clone().into_raw_vec();
+        /// assert_eq!(os_string, OsString::from_raw_vec(os_bytes).unwrap());
+        /// #
+        /// # Ok::<_, io::Error>(())
+        /// ```
+        ///
+        /// [`assert_from_raw_vec`]: Self::assert_from_raw_vec
+        #[cfg_attr(
+            os_str_bytes_docs_rs,
+            doc(cfg(feature = "checked_conversions"))
+        )]
+        fn from_raw_vec(string: Vec<u8>) -> Result<Self>;
+    }
+
+    /// Converts a platform-native string into an equivalent byte string.
     ///
-    /// The returned byte string will use an [unspecified encoding].
+    /// The returned string will use an [unspecified encoding].
     ///
     /// # Examples
     ///
     /// ```
-    /// use std::env;
-    /// # use std::io;
+    /// use std::ffi::OsString;
     ///
     /// use os_str_bytes::OsStringBytes;
     ///
-    /// let os_string = env::current_exe()?;
-    /// println!("{:?}", os_string.into_raw_vec());
-    /// #
-    /// # Ok::<_, io::Error>(())
+    /// let string = "foobar".to_owned();
+    /// let os_string: OsString = string.clone().into();
+    /// assert_eq!(string.into_bytes(), os_string.into_raw_vec());
     /// ```
     ///
     /// [unspecified encoding]: self#encoding
@@ -396,6 +562,11 @@ pub trait OsStringBytes: private::Sealed + Sized {
 }
 
 impl OsStringBytes for OsString {
+    #[inline]
+    fn assert_from_raw_vec(string: Vec<u8>) -> Self {
+        expect_encoded!(imp::os_string_from_vec(string))
+    }
+
     #[inline]
     fn from_raw_vec(string: Vec<u8>) -> Result<Self> {
         imp::os_string_from_vec(string).map_err(EncodingError)
@@ -408,6 +579,11 @@ impl OsStringBytes for OsString {
 }
 
 impl OsStringBytes for PathBuf {
+    #[inline]
+    fn assert_from_raw_vec(string: Vec<u8>) -> Self {
+        OsString::assert_from_raw_vec(string).into()
+    }
+
     #[inline]
     fn from_raw_vec(string: Vec<u8>) -> Result<Self> {
         OsString::from_raw_vec(string).map(Into::into)
@@ -425,7 +601,14 @@ mod private {
     use std::path::Path;
     use std::path::PathBuf;
 
+    if_raw_str! {
+        use std::borrow::Cow;
+
+        use super::RawOsStr;
+    }
+
     pub trait Sealed {}
+
     impl Sealed for char {}
     impl Sealed for OsStr {}
     impl Sealed for OsString {}
@@ -433,6 +616,8 @@ mod private {
     impl Sealed for PathBuf {}
     impl Sealed for &str {}
     impl Sealed for &String {}
-    #[cfg(feature = "deprecated-byte-patterns")]
-    impl Sealed for u8 {}
+
+    if_raw_str! {
+        impl Sealed for Cow<'_, RawOsStr> {}
+    }
 }
